#!/bin/bash

#######################################################
# Setup script for Autolab and Tango for WSUV based on
# Autolab setup script written by CMU
#######################################################

if [[ $# -eq 0 ]]
then
	echo "No argument detected, are you sure you want to continue with default settings? [y/n]"
	read REPLY
	echo    # (optional) move to a new line
	if [[ ! $REPLY =~ ^[Yy]$ ]]
	then
		printf "Some arguments are:"
		printf "\n\t-y - skips all confirmation prompts
	-p [ARG] - Overrides the random generation of MySQL root password with [ARG]
	-u [ARG] - Custom name of user to be created"
		printf "\nSome example usage are:\n
	./setup.sh -yp MySecreteSQLPASS -e dev\n"
		exit 0
	fi	
fi


#######################################
## Initialization and Helper Functions
#######################################
init_helper_vars(){
	if [ -z "$SETUP_PATH" ];
	then 
		if [ ${PWD##*/} = "wsuv" ]; then
			tmp=$(pwd)
			SETUP_PATH="$(dirname $tmp)"
		else
			SETUP_PATH=$(pwd)
		fi
	fi


	if [ -z "$CREATE_USER" ];
	then
		CREATE_USER="autolab"
	fi

	if [ -z "$LOG_FILE" ];
	then
		LOG_FILE=$(mktemp)
	fi
	VERBOSE_LOG="/tmp/setup-system-log"

}

###########################
## Global Helper Functions
###########################

log(){ echo -e "\e[32m$*\e[39m" 2>&1 | tee -a "$LOG_FILE"; }

output(){ echo -e "$*" 2>&1 | tee -a "$LOG_FILE"; }

warn(){ echo -e "\e[33m**WARNING**  $* **WARNING**\e[39m" 2>&1 | tee -a "$LOG_FILE"; }

fail(){ echo -e "\e[31m**ERROR** $* **ERROR**\e[39m" 2>&1 | tee -a "$LOG_FILE"; } 

clean_var(){
	unset AUTOLAB_REPO
	unset LOG_FILE
	unset CREATE_USER
	unset VERBOSE_LOG
	unset MYSQL_ROOT_PW
}
install_if_not_exist(){
	_=$(command -v "$1");
	retval=$?
	if [ "$retval" != "0" ];
	then
		log "installing $1"
		if [ "$1" == "docker" ];
		then
			curl -sSL https://get.docker.com | sh;
		else
			sudo apt -y -qq install "$1"
		fi
	fi
}

## Trap for completion and error
cleanup(){
	EXIT_CODE=$?
	unset AUTOLAB_REPO
	unset LOG_FILE
	unset CREATE_USER
	unset VERBOSE_LOG
	unset MYSQL_ROOT_PW
	exit ${EXIT_CODE:-0}
}

err_report(){
	ERR_CODE=$?
	
	# Ignore Ctrl-C intrrupt
	if [ $ERR_CODE == 130 ];
		then
			return
	fi

	ERR_LINE=$(sed -n "$1p" < "$SETUP_PATH" | sed -e 's/^[ \t]*//')
	warn "Failed command: $ERR_LINE"
	fail "Line $1 of script has return value of $ERR_CODE. The log file is saved at $LOG_FILE."
	exit $ERR_CODE
}

trap 'err_report $LINENO' ERR

## Check if a directory already exists and confirm deletion

dir_check(){
	dir_res=0
	log "Checking for installation of $1"
	if [[ -d "$1" ]];
	then
		warn "$1 directory already exists"
		if [ "$CONFIRM" == "0" ];
		then
			log "Would you like to overwrite $1 directory?[y/n]"
			read -rn1 res
			echo -e "\n"
			if [ "$res" == "y" ];
			then
				warn "Removing old directory"
				rm -rf "$1"
				dir_res=1
			else
				dir_res=0
			fi
		else
			warn "Removing old directory"
			rm -rf "$1"
			dir_res=1
		fi
	else
		dir_res=1
	fi

}

##############################################
## Step 0: Setup prereq
##############################################

prereq_setup(){
	if [ ! -e ~/.ssh ]; then
		warn "No .ssh directory here."
		mkdir ~/.ssh/
		chmod 700 ~/.ssh/
	fi

	if [ ! -e ~/.ssh/authorized_keys ]; then
		warn "No .ssh/authorized_keys file here."
		touch ~/.ssh/authorized_keys
		chmod 644 ~/.ssh/authorized_keys 
	fi
		
	echo "Installing stuff."
	sudo apt install emacs25-nox


	echo "Making users"
	sudo useradd -m -U -s /bin/bash "$CREATE_USER"
	sudo useradd -m -g "$CREATE_USER" -s /bin/bash tango

	USER_HOME="/home/$CREATE_USER"

	sudo mkdir $USER_HOME/.ssh
	sudo cp ~/.ssh/authorized_keys $USER_HOME/.ssh
	sudo chmod 700 $USER_HOME/.ssh
	sudo chmod 600 $USER_HOME/.ssh/authorized_keys
	sudo chown -R $usr:autolab $USER_HOME/.ssh

	echo "setup $CREATE_USER"
	sudo ls -l $USER_HOME/.ssh

	sudo cp ~/.ssh/* $USER_HOME/.ssh/
	sudo find /home/$CREATE_USER/.ssh/ -type f -exec chmod 600 {} \;
	sudo chown -R $CREATE_USER:autolab $USER_HOME/.ssh

	log "Exporting PATH to $USER_HOME"
}

##############################################
## Step 1: Parse Arguments passed to setup.sh
##############################################

parse_args(){
	while getopts "yp:u:" opt;
	do
		case $opt in
			p) 
				warn "MySQL root password provided."
				MYSQL_ROOT_PW=$OPTARG
				;;
			y)
				warn "Skipping all confirmation prompts."
				CONFIRM=1
				;;
			u)
				CREATE_USER=$OPTARG
				log "Will create user: $CREATE_USER"
				;;
			\?)
				fail "Unrecognized option -$OPTARG."
				exit 1
				;;
		esac
	done
}


#############################
## Step 2: Environment Setup
#############################

environment_setup(){
	log "Updating system and installing prerequisites..."
	sudo apt -y -qq update
	sudo apt -y -qq upgrade

	# Rbenv dependencies	
	sudo apt -y -qq install autoconf bison build-essential libssl1.0-dev libyaml-dev libreadline6-dev zlib1g-dev libncurses5-dev libffi-dev libgdbm5 libgdbm-dev
	sudo apt -y -qq install libmysqlclient-dev
	sudo apt -y -qq install libssl-dev
	sudo apt -y -qq install libgmp3-dev


	log "Syncing system time"
	sudo timedatectl set-ntp no
	sudo apt -y -qq install ntp;

    
    # Checkif curl, docker, and screen exist, if not install it
	install_if_not_exist curl;
	install_if_not_exist docker;
	install_if_not_exist screen;
}


#########################################
## Step 3: MySQL and SQLite Installation
#########################################

db_setup(){
	log "Installing SQLite"
	sudo apt -y -qq install sqlite3 libsqlite3-dev
	
	log "Installing MySQL Server"

	if hash mysql 2>/dev/null;
	then
		warn "MySQL Server alread esists on the system"
		log "Do you want to skip configuration?[y/n]"
		read -rn1 res
		if [ "$res" = "y" ];
		then
			read -rs -p "Root password for MySQL Server: " MYSQL_ROOT_PW
			echo -e "\n"
			read -n1 -rsp $'Ensure that MySQL server is running, press any key to continue'
			echo -e "\n"
			return
		else
			log "Attempting to install MySQL Server on top of existing installation"
		fi
	fi

	# Install and configure MySQL
	if [ -z "$MYSQL_ROOT_PW" ];
	then
		MYSQL_ROOT_PW="$(date +%s | sha256sum | base64 | head -c 32)"
		log "Your randomly-generated root password for MySQL server is $MYSQL_ROOT_PW."
	fi
	
	echo "mysql-server mysql-server/root_password password $MYSQL_ROOT_PW" | sudo debconf-set-selections
	echo "mysql-server mysql-server/root_password_again password $MYSQL_ROOT_PW" | sudo debconf-set-selections	
	sudo apt -y -qq install mysql-server


	log "Granting MySQL database permissions ..."

	
	mysql -uroot -p"$MYSQL_ROOT_PW" -e "GRANT ALL PRIVILEGES ON ${CREATE_USER}_development.* TO '$CREATE_USER'@'localhost' IDENTIFIED BY '<password>'"
	mysql -uroot -p"$MYSQL_ROOT_PW" -e "GRANT ALL PRIVILEGES ON ${CREATE_USER}_test.* TO '$CREATE_USER'@'localhost' IDENTIFIED BY '<password>'"
	mysql -uroot -p"$MYSQL_ROOT_PW" -e "GRANT ALL PRIVILEGES ON ${CREATE_USER}_production.* TO '$CREATE_USER'@'%' IDENTIFIED BY '<password>'"
	log "Should be granted"
}


########################
## Step 4: Apache Setup
########################

apache_setup(){
	log "Setting up Apache2 and Passenger"
	sudo apt install -y dirmngr gnupg
	sudo apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv-keys 561F9B9CAC40B2F7
	sudo apt install -y apt-transport-https ca-certificates
	sudo sh -c 'echo deb https://oss-binaries.phusionpassenger.com/apt/passenger bionic main > /etc/apt/sources.list.d/passenger.list'
	sudo apt update
	sudo apt install -y apache2 libapache2-mod-passenger
	sudo a2enmod passenger
	sudo service apache2 restart
}

###########################
## Step 5: Apache2 Config
###########################

apache_config(){
	log "Copying apache configs"
	sudo a2enmod ssl
	sudo cp "$SETUP_PATH"/wsuv/cover/000-default.conf /etc/apache2/sites-available/000-default.conf
	sudo cp "$SETUP_PATH"/wsuv/cover/redirect.conf /etc/apache2/sites-available/000-default-redirect.conf
	sudo cp "$SETUP_PATH"/wsuv/cover/default-ssl.conf /etc/apache2/sites-available/011-autolab-ssl-example.conf
	
	log "Editing apache config"
	sudo sed -i "s/<username>/$CREATE_USER/g" /etc/apache2/sites-available/000-default.conf
	sudo sed -i "s/<username>/$CREATE_USER/g" /etc/apache2/sites-available/011-autolab-ssl-example.conf

	RVER="$(cat $SETUP_PATH/.ruby-version)"
	log "Forcing apache to user ruby $RVER"
	sudo sed -i "s/<RVER>/$RVER/g" /etc/apache2/sites-available/000-default.conf
	sudo sed -i "s/<RVER>/$RVER/g" /etc/apache2/sites-available/011-autolab-ssl-example.conf

	unset RVER
	sudo service apache2 restart
}

###########################
## Step 6: Finish Install
###########################
copy_autolab_over(){
    USER_HOME="/home/$CREATE_USER" # For some reason, the variable keeps emptying out, so we're re-value it
	log "Checking $USER_HOME for existing Autolab"
	dir_check "$USER_HOME/Autolab"
	if [ "$dir_res" == "1" ];
	then
		log "Moving directory to $USER_HOME"

		sudo cp -r $SETUP_PATH $USER_HOME/.
		sudo chown -R $CREATE_USER $USER_HOME/Autolab
	else
		warn "Skipping cloning Autolab repository."
		cd ~/Autolab || exit; AUTOLAB_PATH=$(pwd)
	fi
}



###########################
## Step 7: Finish Install
###########################

fin(){
	log "Installation finished, please log out and back in for all changes to take effect"
	log "Your SQL root password is $MYSQL_ROOT_PW"
	log "See INSTALLATION_GUIDE.txt on instructions on how to run Tango"
	log "Check $LOG_FILE AND $VERBOSE_LOG for log of the installation"

	printf "\nFinished setting up prerequirements!
	To Set up Autolab front-end:
	1.\tLog into the user, autolab
	2.\tRun the setup.sh script
"
	trap 'cleanup' EXIT
}

clean_var
init_helper_vars
prereq_setup 2>&1 | tee $VERBOSE_LOG
parse_args "$@" 2>&1 | tee -a $VERBOSE_LOG
environment_setup 2>&1 | tee -a $VERBOSE_LOG
db_setup 2>&1 | tee -a $VERBOSE_LOG
apache_setup 2>&1 | tee -a $VERBOSE_LOG
apache_config 2>&1 | tee -a $VERBOSE_LOG
copy_autolab_over 2>&1 | tee -a $VERBOSE_LOG
fin 2>&1 | tee -a $VERBOSE_LOG


