#!/bin/bash

#######################################################
# Setup script for Autolab and Tango for WSUV based on
# Autolab setup script written by CMU
#######################################################


#######################################
## Initialization and Helper Functions
#######################################
init_helper_vars(){
	if [ -z "$SETUP_PATH" ];
	then
		if [ ${PWD##*/} = "wsuv" ]; then
			tmp=$(pwd)
			SETUP_PATH="$(dirname $tmp)"
		else
			SETUP_PATH=$(pwd)
		fi
	fi

	if [ -z "$RBENV_REPO" ];
	then
		RBENV_REPO="https://github.com/rbenv/rbenv.git"
	fi

	if [ -z "$RBUILD_REPO" ];
	then
		RBUILD_REPO="https://github.com/rbenv/ruby-build.git"
	fi

	if [ -z "$LOG_FILE" ];
	then
		LOG_FILE=$(mktemp)
	fi
	if [ -z "$VERBOSE_LOG" ];
	then
		VERBOSE_LOG="/tmp/setup-log"
	fi
}


###########################
## Global Helper Functions
###########################

log(){ echo -e "\e[32m$*\e[39m" 2>&1 | tee -a "$LOG_FILE"; }

output(){ echo -e "$*" 2>&1 | tee -a "$LOG_FILE"; }

warn(){ echo -e "\e[33m**WARNING**  $* **WARNING**\e[39m" 2>&1 | tee -a "$LOG_FILE"; }

fail(){ echo -e "\e[31m**ERROR** $* **ERROR**\e[39m" 2>&1 | tee -a "$LOG_FILE"; } 

clean_var(){
	unset LOG_FILE
	unset MYSQL_ROOT_PW
}
install_if_not_exist(){
	_=$(command -v "$1");
	retval=$?
	if [ "$retval" != "0" ];
	then
		log "installing $1"
		if [ "$1" == "docker" ];
		then
			curl -sSL https://get.docker.com | sh;
		else
			sudo apt -y -qq install "$1"
		fi
	fi
}

## Trap for completion and error
cleanup(){
	EXIT_CODE=$?
	unset LOG_FILE
	unset MYSQL_ROOT_PW
	unset VERBOSE_LOG
	exit ${EXIT_CODE:-0}
}

err_report(){
	ERR_CODE=$?
	
	# Ignore Ctrl-C intrrupt
	if [ $ERR_CODE == 130 ];
		then
			return
	fi

	ERR_LINE=$(sed -n "$1p" < "$SETUP_PATH" | sed -e 's/^[ \t]*//')
	warn "Failed command: $ERR_LINE"
	fail "Line $1 of script has return value of $ERR_CODE. The log file is saved at $LOG_FILE."
	exit $ERR_CODE
}

trap 'err_report $LINENO' ERR

## Check if a directory already exists and confirm deletion

dir_check(){
	dir_res=0
	log "Checking for installation of $1"
	if [[ -d "$1" ]];
	then
		warn "$1 directory already exists"
		if [ "$CONFIRM" == "0" ];
		then 
			log "Would you like to overwrite $1 directory?[y/n]"
			read -rn1 res
			echo -e "\n"
			if [ "$res" == "y" ];
			then
				warn "Removing old directory"
				rm -rf "$1"
				dir_res=1
			else
				dir_res=0
			fi
		else
			warn "Removing old directory"
			rm -rf "$1"
			dir_res=1
		fi
	else
		dir_res=1
	fi

}

##############################################
## Step 1: Parse Arguments passed to setup.sh
##############################################

parse_args(){
	while getopts "y:" opt;
	do
		case $opt in
			y)
				warn "Skipping all confirmation prompts."
				CONFIRM=1
				;;
			\?)
				fail "Unrecognized option -$OPTARG."
				exit 1
				;;
		esac
	done
}

#########################
## Step 2: Set up Rbenv
#########################
rbenv_setup(){
	cd ~/Autolab
	rb_path="$HOME"/.rbenv
		
	if [ ! -d $rb_path ]
	then
		log "Cloning Rbenv repo from Github to ~/.rbenv"
			
		git clone $RBENV_REPO ~/.rbenv

		log "Setting up Rbenv"
				
		echo 'export PATH="$HOME/.rbenv/bin:$PATH"' >> ~/.bashrc
		echo 'eval "$(rbenv init -)"' >> ~/.bashrc

		log "Adding ruby-build to rbenv"
		git clone $RBUILD_REPO ~/.rbenv/plugins/ruby-build

		export PATH="$HOME/.rbenv/bin:$PATH"
		ruby_setup  2>&1 | tee -a $VERBOSE_LOG
	else
		warn "rbenv detected, skipping rbenv setup."
	fi
}

########################
## Step 3: Install Ruby (If first time)
########################

ruby_setup(){
	RVER=$(grep "rbenv/versions" /etc/apache2/sites-available/000-default.conf | sed -Ee "s/.*versions\/([0-9\.]+)\/bin\/ruby/\1/")
	REPO_VER=$(cat $SETUP_PATH/.ruby-version)
	if [[ "$RVER" != "$REPO_VER" ]];
	then
		fail  "Apache's ruby version ($RVER) does not match this repo's version ($REPO_VER)"
		trap 'cleanup' EXIT
		exit 1
	fi
	# Using apache's require ruby instead of "$SETUP_PATH"/.ruby-version
	log "Installing ruby $RVER"
	rbenv install "$RVER"
	rbenv global "$RVER"
	output "Using ruby version: $(ruby -v)"
	unset RVER
}

##############################
## Step 4: Rails Installation
##############################

rails_setup(){
	output "Using ruby version: $(ruby -v)"
	log "Installing necessary gems for Autolab"
	cd "$SETUP_PATH" || exit
	echo $(pwd)
	gem install bundler -v 1.17.3
	# ignore gem's post install message for cleaner verbose and log
	bundle config --global ignore_messages true
	bundle install
	rbenv rehash
}

######################################
## Step 5: Initialize Autolab configs
######################################

autolab_setup(){

	cd "$SETUP_PATH" || exit
	log "Initializing Autolab configurations"
	cp "$SETUP_PATH"/wsuv/cover/modified-database.yml "$SETUP_PATH"/config/database.yml
	sed -i "s/<username>/$USER/g" "$SETUP_PATH"/config/database.yml

	cp "$SETUP_PATH"/config/school.yml.template "$SETUP_PATH"/config/school.yml

	cp "$SETUP_PATH"/config/initializers/devise.rb.template "$SETUP_PATH"/config/initializers/devise.rb

	sed -i "s/<YOUR-SECRET-KEY>/$(bundle exec rake secret)/g" "$SETUP_PATH"/config/initializers/devise.rb

	cp "$SETUP_PATH"/config/autogradeConfig.rb.template "$SETUP_PATH"/config/autogradeConfig.rb
}

###########################################
## Step 6: Autloab database initialization
###########################################

autolab_init(){

    log "Generating (raking) tables in Autolab database"
	cd "$SETUP_PATH" || exit
	output "Using ruby version: $(ruby -v)"
	rake db:create:all
	rake db:reset
	rake db:migrate
	# cp "$SETUP_PATH"/wsuv/cover/configure_users.rb "$SETUP_PATH"/.
	# ./bin/rails r configure_users.rb development # Run this to generate some testing user, like admin@foo.bar
}


#############################################
## Step 7: Get Autolab Ready for Deployment
#############################################
autolab_deploy(){

	log "Getting Autolab ready for deployment"
	if [ -z "$NUMPROC" ];
	then
			NUMPROC=$(grep -c ^processor /proc/cpuinfo)
	fi
			
	bundle install --deployment --without development test --jobs "$NUMPROC"
	rake assets:precompile RAILS_ENV=production
	
	cd "$SETUP_PATH"/config/environments/ || exit
	cp production.template.rb production.rb
}

###########################
## Step 8: Finish Install
###########################

fin(){
	log "Installation finished, please log out and back in for all changes to take effect"
	log "See INSTALLATION_GUIDE.txt on instructions on how to run Tango"
	log "Check $LOG_FILE AND $VERBOSE_LOG for log of the installation"

	echo "Autolab is now installed, and should be accessible via http."
	warn "If you want to run via https and/or in the production environment, additional steps will be needed."
	echo
	warn "> Be sure that the Autolab instance is pointed towards your Tango install correctly <"

	trap 'cleanup' EXIT
}

clean_var
init_helper_vars
parse_args "$@" 2>&1 | tee $VERBOSE_LOG
rbenv_setup 2>&1 | tee -a $VERBOSE_LOG

# Refresh the shell
export PATH="$HOME/.rbenv/bin:$PATH"
eval "$(rbenv init -)"
. ~/.bashrc

rails_setup 2>&1 | tee -a $VERBOSE_LOG
autolab_setup 2>&1 | tee -a $VERBOSE_LOG
autolab_init 2>&1 | tee -a $VERBOSE_LOG
autolab_deploy 2>&1 | tee -a $VERBOSE_LOG
fin 2>&1 | tee -a $VERBOSE_LOG
